@@ -1,4 +1,4 @@
-var should = require("should"); // eslint-disable-line
+require("should");
 var path = require("path");
 var runLoaders = require("../").runLoaders;
 var getContext = require("../").getContext;
@@ -11,7 +11,7 @@ describe("runLoaders", function() {
 			resource: path.resolve(fixtures, "resource.bin")
 		}, function(err, result) {
 			if(err) return done(err);
-			result.result.should.be.eql([new Buffer("resource", "utf-8")]); // eslint-disable-line
+			result.result.should.be.eql([Buffer.from("resource", "utf-8")]);
 			result.cacheable.should.be.eql(true);
 			result.fileDependencies.should.be.eql([
 				path.resolve(fixtures, "resource.bin")
@@ -114,6 +114,45 @@ describe("runLoaders", function() {
 			done();
 		});
 	});
+	it("should interpret explicit `undefined` values from async 'pitch' loaders", function(done) {
+		runLoaders({
+			resource: path.resolve(fixtures, "resource.bin"),
+			loaders: [
+				path.resolve(fixtures, "simple-loader.js"),
+				path.resolve(fixtures, "pitch-async-undef-loader.js"),
+				path.resolve(fixtures, "pitch-promise-undef-loader.js")
+			]
+		}, function(err, result) {
+			if(err) return done(err);
+			result.result.should.be.eql([
+				"resource-simple"
+			]);
+			result.cacheable.should.be.eql(true);
+			result.fileDependencies.should.be.eql([
+				path.resolve(fixtures, "resource.bin")
+			]);
+			result.contextDependencies.should.be.eql([]);
+			done();
+		});
+	});
+	it("should interrupt pitching when async loader completes with any additional non-undefined values", function(done) {
+		runLoaders({
+			resource: path.resolve(fixtures, "resource.bin"),
+			loaders: [
+				path.resolve(fixtures, "simple-loader.js"),
+				path.resolve(fixtures, "pitch-async-undef-some-loader.js")
+			]
+		}, function(err, result) {
+			if(err) return done(err);
+			result.result.should.be.eql([
+				"undefined-simple"
+			]);
+			result.cacheable.should.be.eql(true);
+			result.fileDependencies.should.be.eql([]);
+			result.contextDependencies.should.be.eql([]);
+			done();
+		});
+	});
 	it("should be possible to add dependencies", function(done) {
 		runLoaders({
 			resource: path.resolve(fixtures, "resource.bin"),
@@ -125,13 +164,14 @@ describe("runLoaders", function() {
 			result.cacheable.should.be.eql(true);
 			result.fileDependencies.should.be.eql(["a", "b"]);
 			result.contextDependencies.should.be.eql(["c"]);
-			result.result.should.be.eql(["resource\n" + JSON.stringify(["a", "b"]) + JSON.stringify(["c"])]);
+			result.missingDependencies.should.be.eql(["d"]);
+			result.result.should.be.eql(["resource\n" + JSON.stringify(["a", "b"]) + JSON.stringify(["c"]) + JSON.stringify(["d"])]);
 			done();
 		});
 	});
 	it("should have to correct keys in context", function(done) {
 		runLoaders({
-			resource: path.resolve(fixtures, "resource.bin") + "?query",
+			resource: path.resolve(fixtures, "resource.bin") + "?query#frag",
 			loaders: [
 				path.resolve(fixtures, "keys-loader.js") + "?loader-query",
 				path.resolve(fixtures, "simple-loader.js")
@@ -141,25 +181,27 @@ describe("runLoaders", function() {
 			try {
 				JSON.parse(result.result[0]).should.be.eql({
 					context: fixtures,
-					resource: path.resolve(fixtures, "resource.bin") + "?query",
+					resource: path.resolve(fixtures, "resource.bin") + "?query#frag",
 					resourcePath: path.resolve(fixtures, "resource.bin"),
 					resourceQuery: "?query",
+					resourceFragment: "#frag",
 					loaderIndex: 0,
 					query: "?loader-query",
 					currentRequest: path.resolve(fixtures, "keys-loader.js") + "?loader-query!" +
 						path.resolve(fixtures, "simple-loader.js") + "!" +
-						path.resolve(fixtures, "resource.bin") + "?query",
+						path.resolve(fixtures, "resource.bin") + "?query#frag",
 					remainingRequest: path.resolve(fixtures, "simple-loader.js") + "!" +
-						path.resolve(fixtures, "resource.bin") + "?query",
+						path.resolve(fixtures, "resource.bin") + "?query#frag",
 					previousRequest: "",
 					request: path.resolve(fixtures, "keys-loader.js") + "?loader-query!" +
 						path.resolve(fixtures, "simple-loader.js") + "!" +
-						path.resolve(fixtures, "resource.bin") + "?query",
+						path.resolve(fixtures, "resource.bin") + "?query#frag",
 					data: null,
 					loaders: [{
 						request: path.resolve(fixtures, "keys-loader.js") + "?loader-query",
 						path: path.resolve(fixtures, "keys-loader.js"),
 						query: "?loader-query",
+						fragment: "",
 						data: null,
 						pitchExecuted: true,
 						normalExecuted: true
@@ -167,6 +209,7 @@ describe("runLoaders", function() {
 						request: path.resolve(fixtures, "simple-loader.js"),
 						path: path.resolve(fixtures, "simple-loader.js"),
 						query: "",
+						fragment: "",
 						data: null,
 						pitchExecuted: true,
 						normalExecuted: true
@@ -196,6 +239,7 @@ describe("runLoaders", function() {
 					resource: path.resolve(fixtures, "resource.bin") + "?query",
 					resourcePath: path.resolve(fixtures, "resource.bin"),
 					resourceQuery: "?query",
+					resourceFragment: "",
 					loaderIndex: 0,
 					query: {
 						ident: "ident",
@@ -212,6 +256,7 @@ describe("runLoaders", function() {
 						request: path.resolve(fixtures, "keys-loader.js") + "??ident",
 						path: path.resolve(fixtures, "keys-loader.js"),
 						query: "??ident",
+						fragment: "",
 						options: {
 							ident: "ident",
 							loader: "query"
@@ -241,7 +286,7 @@ describe("runLoaders", function() {
 			done();
 		});
 	});
-	it("should process omit BOM on string convertion", function(done) {
+	it("should process omit BOM on string conversion", function(done) {
 		runLoaders({
 			resource: path.resolve(fixtures, "bom.bin"),
 			loaders: [
@@ -279,6 +324,7 @@ describe("runLoaders", function() {
 						request: path.resolve(fixtures, "identity-loader.js"),
 						path: path.resolve(fixtures, "identity-loader.js"),
 						query: "",
+						fragment: "",
 						data: {
 							identity: true
 						},
@@ -288,6 +334,7 @@ describe("runLoaders", function() {
 						request: path.resolve(fixtures, "keys-loader.js"),
 						path: path.resolve(fixtures, "keys-loader.js"),
 						query: "",
+						fragment: "",
 						data: null,
 						pitchExecuted: true,
 						normalExecuted: true
@@ -317,6 +364,7 @@ describe("runLoaders", function() {
 					resource: "?query",
 					resourcePath: "",
 					resourceQuery: "?query",
+					resourceFragment: "",
 					loaderIndex: 0,
 					query: {
 						ok: true
@@ -331,6 +379,56 @@ describe("runLoaders", function() {
 						request: path.resolve(fixtures, "keys-loader.js") + "??my-ident",
 						path: path.resolve(fixtures, "keys-loader.js"),
 						query: "??my-ident",
+						fragment: "",
+						ident: "my-ident",
+						options: {
+							ok: true
+						},
+						data: null,
+						pitchExecuted: true,
+						normalExecuted: true
+					}]
+				});
+			} catch(e) {
+				return done(e);
+			}
+			done();
+		});
+	});
+	it("should have to correct keys in context with only resource fragment", function(done) {
+		runLoaders({
+			resource: "#fragment",
+			loaders: [{
+				loader: path.resolve(fixtures, "keys-loader.js"),
+				options: {
+					ok: true
+				},
+				ident: "my-ident"
+			}]
+		}, function(err, result) {
+			if(err) return done(err);
+			try {
+				JSON.parse(result.result[0]).should.be.eql({
+					context: null,
+					resource: "#fragment",
+					resourcePath: "",
+					resourceQuery: "",
+					resourceFragment: "#fragment",
+					loaderIndex: 0,
+					query: {
+						ok: true
+					},
+					currentRequest: path.resolve(fixtures, "keys-loader.js") + "??my-ident!#fragment",
+					remainingRequest: "#fragment",
+					previousRequest: "",
+					request: path.resolve(fixtures, "keys-loader.js") + "??my-ident!" +
+						"#fragment",
+					data: null,
+					loaders: [{
+						request: path.resolve(fixtures, "keys-loader.js") + "??my-ident",
+						path: path.resolve(fixtures, "keys-loader.js"),
+						query: "??my-ident",
+						fragment: "",
 						ident: "my-ident",
 						options: {
 							ok: true
@@ -362,7 +460,7 @@ describe("runLoaders", function() {
 			done();
 		});
 	});
-	it("should return dependencies even if resource is missing", function(done) {
+	it("should return dependencies when resource not found", function(done) {
 		runLoaders({
 			resource: path.resolve(fixtures, "missing.txt"),
 			loaders: [
@@ -378,13 +476,86 @@ describe("runLoaders", function() {
 			done();
 		});
 	});
-	it("should return dependencies even if loader is failing", function(done) {
+	it("should not return dependencies when loader not found", function(done) {
 		runLoaders({
 			resource: path.resolve(fixtures, "resource.bin"),
 			loaders: [
-				path.resolve(fixtures, "failing-loader.js")
+				path.resolve(fixtures, "does-not-exist-loader.js")
 			]
 		}, function(err, result) {
+			err.should.be.instanceOf(Error);
+			err.code.should.be.eql("MODULE_NOT_FOUND");
+			err.message.should.match(/does-not-exist-loader\.js\'($|\n)/i);
+			result.should.be.eql({
+				cacheable: false,
+				fileDependencies: [],
+				contextDependencies: [],
+				missingDependencies: []
+			});
+			done();
+		});
+	});
+	it("should not return dependencies when loader is empty object", function(done) {
+		runLoaders({
+			resource: path.resolve(fixtures, "resource.bin"),
+			loaders: [
+				path.resolve(fixtures, "module-exports-object-loader.js")
+			]
+		}, function(err, result) {
+			err.should.be.instanceOf(Error);
+			err.message.should.match(/module-exports-object-loader.js' is not a loader \(must have normal or pitch function\)$/);
+			result.should.be.eql({
+				cacheable: false,
+				fileDependencies: [],
+				contextDependencies: [],
+				missingDependencies: []
+			});
+			done();
+		});
+	});
+	it("should not return dependencies when loader is otherwise invalid (string)", function(done) {
+		runLoaders({
+			resource: path.resolve(fixtures, "resource.bin"),
+			loaders: [
+				path.resolve(fixtures, "module-exports-string-loader.js")
+			]
+		}, function(err, result) {
+			err.should.be.instanceOf(Error);
+			err.message.should.match(/module-exports-string-loader.js' is not a loader \(export function or es6 module\)$/);
+			result.should.be.eql({
+				cacheable: false,
+				fileDependencies: [],
+				contextDependencies: [],
+				missingDependencies: []
+			});
+			done();
+		});
+	});
+	it("should return dependencies when loader throws error", function(done) {
+		runLoaders({
+			resource: path.resolve(fixtures, "resource.bin"),
+			loaders: [
+				path.resolve(fixtures, "throws-error-loader.js")
+			]
+		}, function(err, result) {
+			err.should.be.instanceOf(Error);
+			err.message.should.match(/^resource$/i);
+			result.fileDependencies.should.be.eql([
+				path.resolve(fixtures, "resource.bin")
+			]);
+			done();
+		});
+	});
+	it("should return dependencies when loader rejects promise", function(done) {
+		var once = true;
+		runLoaders({
+			resource: path.resolve(fixtures, "resource.bin"),
+			loaders: [
+				path.resolve(fixtures, "promise-error-loader.js")
+			]
+		}, function(err, result) {
+			if(!once) return done(new Error("should not be called twice"));
+			once = false;
 			err.should.be.instanceOf(Error);
 			err.message.should.match(/^resource$/i);
 			result.fileDependencies.should.be.eql([
@@ -436,16 +607,101 @@ describe("runLoaders", function() {
 		});
 		delete global.System;
 	});
+	if(+process.versions.modules >= 83) {
+		it("should load a loader using import()", function(done) {
+			runLoaders({
+				resource: path.resolve(fixtures, "resource.bin"),
+				loaders: [
+					{
+						loader: path.resolve(fixtures, "esm-loader.mjs"),
+						type: "module"
+					}
+				]
+			}, function(err, result) {
+				if(err) return done(err);
+				result.result.should.be.eql(["resource-esm"]);
+				result.cacheable.should.be.eql(true);
+				result.fileDependencies.should.be.eql([
+					path.resolve(fixtures, "resource.bin")
+				]);
+				result.contextDependencies.should.be.eql([]);
+				done();
+			});
+		});
+	}
+	it("should support escaping in resource", function(done) {
+		runLoaders({
+			resource: path.resolve(fixtures, "res\0#ource.bin")
+		}, function(err, result) {
+			if(err) return done(err);
+			result.result.should.be.eql([Buffer.from("resource", "utf-8")]);
+			result.cacheable.should.be.eql(true);
+			result.fileDependencies.should.be.eql([
+				path.resolve(fixtures, "res#ource.bin")
+			]);
+			result.contextDependencies.should.be.eql([]);
+			done();
+		});
+	});
+	it("should have to correct keys in context when using escaping", function(done) {
+		runLoaders({
+			resource: path.resolve(fixtures, "res\0#ource.bin") + "?query\0#frag",
+			loaders: [
+				path.resolve(fixtures, "keys-loader.js") + "?loader\0#query"
+			]
+		}, function(err, result) {
+			if(err) return done(err);
+			try {
+				JSON.parse(result.result[0]).should.be.eql({
+					context: fixtures,
+					resource: path.resolve(fixtures, "res\0#ource.bin") + "?query\0#frag",
+					resourcePath: path.resolve(fixtures, "res#ource.bin"),
+					resourceQuery: "?query#frag",
+					resourceFragment: "",
+					loaderIndex: 0,
+					query: "?loader#query",
+					currentRequest: path.resolve(fixtures, "keys-loader.js") + "?loader\0#query!" +
+						path.resolve(fixtures, "res\0#ource.bin") + "?query\0#frag",
+					remainingRequest: path.resolve(fixtures, "res\0#ource.bin") + "?query\0#frag",
+					previousRequest: "",
+					request: path.resolve(fixtures, "keys-loader.js") + "?loader\0#query!" +
+						path.resolve(fixtures, "res\0#ource.bin") + "?query\0#frag",
+					data: null,
+					loaders: [{
+						request: path.resolve(fixtures, "keys-loader.js") + "?loader\0#query",
+						path: path.resolve(fixtures, "keys-loader.js"),
+						query: "?loader#query",
+						fragment: "",
+						data: null,
+						pitchExecuted: true,
+						normalExecuted: true
+					}]
+				});
+			} catch(e) {
+				return done(e);
+			}
+			done();
+		});
+	});
+
 	describe("getContext", function() {
 		var TESTS = [
 			["/", "/"],
 			["/path/file.js", "/path"],
+			["/path/file.js#fragment", "/path"],
+			["/path/file.js?query", "/path"],
+			["/path/file.js?query#fragment", "/path"],
+			["/path/\0#/file.js", "/path/#"],
 			["/some/longer/path/file.js", "/some/longer/path"],
 			["/file.js", "/"],
 			["C:\\", "C:\\"],
 			["C:\\file.js", "C:\\"],
 			["C:\\some\\path\\file.js", "C:\\some\\path"],
 			["C:\\path\\file.js", "C:\\path"],
+			["C:\\path\\file.js#fragment", "C:\\path"],
+			["C:\\path\\file.js?query", "C:\\path"],
+			["C:\\path\\file.js?query#fragment", "C:\\path"],
+			["C:\\path\\\0#\\file.js", "C:\\path\\#"],
 		];
 		TESTS.forEach(function(testCase) {
 			it("should get the context of '" + testCase[0] + "'", function() {
